#include <stdio.h>
#include <math.h>
#include <avs/flow.h>
#include <avs/avs_data.h>
#include <avs/field.h>
#include <time.h>

extern float implied_vol();

interface()
{
	int out_port;		/* temporary to hold the port number */
	int parm;		/* temporary to hold the parm number */

	/* Set the module name and type */
	AVSset_module_name("Input_interface_IBM", MODULE_DATA);

	/* There are no input ports for this module */

	/* Create output ports for the result */
	out_port = AVScreate_output_port("Field Output", "field 1D scalar uniform float");
	out_port = AVScreate_output_port("Psi", "real");
	out_port = AVScreate_output_port("Rho", "real");
	out_port = AVScreate_output_port("V_bs", "real");
	out_port = AVScreate_output_port("V_ac", "real");
	out_port = AVScreate_output_port("V_es", "real");
	out_port = AVScreate_output_port("V_as", "real");

	/* Add parameter:  */
	parm = AVSadd_parameter("Read Stock Data Browser", "string", 0, 0, 0);
	AVSconnect_widget(parm,"browser");
	/* Add parameter:  */
	parm = AVSadd_parameter("Read Model Parameter Browser", "string", 0, 0, 0);
	AVSconnect_widget(parm,"browser");

	/* Add parameter:  */
	parm = AVSadd_float_parameter("Psi", 1.0, 0.0, 10.0);
	parm = AVSadd_float_parameter("Rho", 0.5, -2.0, 2.0);
	parm = AVSadd_float_parameter("V_BS", 1.0, 0.0, 5.0);
	parm = AVSadd_float_parameter("V_AMC", 1.0, 0.0, 5.0);
	parm = AVSadd_float_parameter("V_EUS", 1.0, 0.0, 5.0);
	parm = AVSadd_float_parameter("V_AMS", 1.0, 0.0, 5.0);

/*	parm = AVSadd_parameter("points/frame","integer", 30, 1, 100);
	AVSconnect_widget(parm,"idial"); */

	/* Add parameter: an enable/disable toggle for the scope */
	parm = AVSadd_parameter("sleep", "boolean", 1, 0, 1);
	parm = AVSadd_parameter("step", "oneshot", 1, 0, 1);
	parm = AVSadd_parameter("optimize", "boolean", 0, 0, 1);
        parm = AVSadd_parameter("abort", "oneshot", 0, 0, 1);

	/* There is no compute function for this module */
}

#define Stockdata 7
#define COMMAND(a) sprintf(inbuf, a); \
  result=AVScommand("kernel", inbuf, &outbuf, &errbuf); \
  if (!result) {AVSerror(errbuf); return(0);}

main(argc,argv)
int	argc;
char	*argv[];
{
   int fin();
   int sleep,step,opt,abort;
   float *v_bs,*v_amc,*v_eus,*v_ams,*psi, *rho;
   char *StockFileName, *ParameterFileName;
   AVSfield_float *output, *input;
   
   int dims[1];
   
   FILE *sfp, *pfp;
   int flag;
   float sz,szdiv,k,rate,t,tdiv,div,call;
   float vt_bs,vt_zero;
   long int total;

   char *outbuf, *errbuf, inbuf[1024];
   int result;

   /* coroutine initilization */
   AVScorout_init(argc,argv,interface);
   AVScorout_set_sync(0); 
   dims[0] = Stockdata;

   while (1) {
     StockFileName = ParameterFileName = NULL;
     total = 0;
     sleep = step = flag = 1;
     input = NULL;
     AVSparameter_visible("Read Stock Data Browser",1);
     AVSparameter_visible("Read Model Parameter Browser",1);

     do {
       AVScorout_wait();
       AVScorout_input(&StockFileName, &ParameterFileName, &psi, &rho,
		       &v_bs,&v_amc,&v_eus,&v_ams,&sleep,&step,&opt,&abort);
       fprintf(stderr,"Stock file = %s Parameter file = %s\n",
	       StockFileName,ParameterFileName);
       if (abort) {
         COMMAND("quit -dialog");
/*         COMMAND("net_clear");
         exit(0); */
       }
     } while (!(StockFileName && ParameterFileName));
     
     if ((sfp = fopen(StockFileName, "r")) == NULL) {
       AVSerror("finance: cannot open %s\n",StockFileName);
       continue;
     }
     
     if ((pfp = fopen(ParameterFileName, "r")) == NULL) {
       AVSerror("finance: cannot open %s\n",ParameterFileName);
       continue;
     }
     
     if (fscanf(pfp, "%f %f",psi, rho) == EOF) {
       AVSerror("finance: incorrect parameter file: %s\n",
		ParameterFileName);
       continue;
     } 
     AVSmodify_float_parameter("Psi",AVS_VALUE, *psi, NULL, NULL);
     AVSmodify_float_parameter("Rho",AVS_VALUE, *rho, NULL, NULL);

     AVSparameter_visible("Read Stock Data Browser",0);
     AVSparameter_visible("Read Model Parameter Browser",0);

     /* Allocate space for stock data */
     input = (AVSfield_float *)AVSdata_alloc
       ("field 1D scalar uniform float",dims);

     while (1) {
       
       if (sleep) AVScorout_wait();
       
       AVScorout_input(&StockFileName, &ParameterFileName, &psi, &rho,
		       &v_bs,&v_amc,&v_eus,&v_ams,&sleep,&step,&opt,&abort);
       if (abort) {
         COMMAND("quit -dialog");
/*         COMMAND("net_clear");
         exit(0); */
       }
       
       if (AVSparameter_changed("optimize")) {
	 AVSparameter_visible("V_BS",(opt==0));
	 AVSparameter_visible("V_AMC",(opt==0));
	 AVSparameter_visible("V_EUS",(opt==0));
 	 AVSparameter_visible("V_AMS",(opt==0));
       }      
       if (get_data(sfp, &sz, &k, &rate, &t, &tdiv, &div, &flag,&call,
		    &vt_bs,&call) == 0) break;
       
       if (flag == 1) { 
	 if (opt==0){
	   AVSmodify_float_parameter("V_BS",AVS_VALUE,vt_bs,NULL,NULL);	 
	   AVSmodify_float_parameter("V_AMC",AVS_VALUE,vt_bs,NULL,NULL);
	   AVSmodify_float_parameter("V_EUS",AVS_VALUE,vt_bs,NULL,NULL);
	   AVSmodify_float_parameter("V_AMS",AVS_VALUE,vt_bs,NULL,NULL);
	   /* simply assuming all model volatilities are the same as BS model's */
	   *v_ams = *v_eus = *v_amc = *v_bs  = vt_bs; 
	 }
	 else {
	   sz = -sz;
	   *v_ams = *v_eus = *v_amc = 0.0;
	   *v_bs  = vt_bs;
	 }
       }
       input->data[0] = sz; input->data[1] = k; input->data[2] = rate;
       input->data[3] = t; input->data[4] = tdiv; input->data[5] = div; 
       input->data[6] = call;
       
       output = input;
       AVScorout_exec(); /* syncronization */
       
       fprintf(stderr,"\n%4li:",++total);
       AVScorout_output(output,psi,rho,v_bs,v_amc,v_eus,v_ams);  
     }
     fprintf(stderr, "\n!!!!!!!! Done !!!!!!!!\n");
     fclose(sfp); fclose(pfp);
     AVSfield_free(input);

     /* initilize parameter  */
     AVSmodify_parameter("Read Stock Data Browser",AVS_VALUE, NULL, 0, 0);
     AVSmodify_parameter("Read Model Parameter Browser", AVS_VALUE, NULL, 0, 0);
     AVSmodify_float_parameter("Psi",AVS_VALUE, 1.0, 0.0, 10.0);
     AVSmodify_float_parameter("Rho",AVS_VALUE, 0.5, -2.0, 2.0);
     AVSmodify_float_parameter("V_BS",AVS_VALUE,1.0,NULL,NULL);
     AVSmodify_float_parameter("V_AMC",AVS_VALUE,1.0,NULL,NULL);
     AVSmodify_float_parameter("V_EUS",AVS_VALUE,1.0,NULL,NULL);
     AVSmodify_float_parameter("V_AMS",AVS_VALUE,1.0,NULL,NULL);

     AVSmodify_parameter("sleep",AVS_VALUE, 1, 0, 1);
     AVSmodify_parameter("step", AVS_VALUE, 1, 0, 1);
     AVSmodify_parameter("optimize", AVS_VALUE, 0, 0, 1);
     AVSmodify_parameter("abort", AVS_VALUE, 0, 0, 1);
     AVSparameter_visible("V_BS",1);
     AVSparameter_visible("V_AMC",1);
     AVSparameter_visible("V_EUS",1);
     AVSparameter_visible("V_AMS",1);
   }
 }

/* read primitive stock data from file and calculate volatilities for models */
int get_data(fp, sz, k, rate, t, tdiv, div, flag, call, v_bs)
FILE *fp;
float *sz, *k, *rate, *t, *tdiv, *div, *call, *v_bs;
int *flag;
{
  float bid,ask,diva[4],tdiva[4];
  char name[4],rstr1[8],rstr2[8];
  int time,vol,tcall,mature;
  long int date,exp;
  
  static float oldrate;
  int strange = 0;
  
  while (fscanf(fp,"%li %s %i%li%i%f%f%f%f%i%i%i %s %s %f%f%f%f%f%f",
		&date,name,&time,&exp,&mature,k,
		&bid,&ask,sz,&vol,&tcall,flag,rstr1,rstr2,
		&tdiva[1],&diva[1],&tdiva[2],&diva[2],&tdiva[3],&diva[3]) != EOF) 
    {
      if (mature == 0.0) continue;
      /* Some of the records have "." in place of interest rates: */
      if (rstr2[0] != '.')
	sscanf(rstr1,"%f",&oldrate);
      
      *rate = oldrate/100.0;     /* Interest rates are given in percent */
      *call = 0.5*(bid + ask);  /* Average bid and ask to get call price */
      *div = diva[1];
      *tdiv = tdiva[1]/365.0;
      *t = mature/365.0;
      
      /* Some records have strike + call <= stock, which is impossible. 
	 Ignore those */
      
      if ((tdiva[2]!=0.0)||(diva[2]!=0.0)||((*flag == 1) && (*k + *call <= *sz))
	  ||((*flag==0) && (strange==1))) {
	fprintf(stderr,"Input data error\n");
	continue;
      }
      
      if (*flag == 1) {
	fprintf(stderr,"\ncalcualting volitility ...\n");
	if ((*v_bs = implied_vol(*sz,*k,*rate,*t,*call,*div,*tdiv)) <= 0.0)
	  continue;
      }
      return(1);
    };
    return(0);  /* end of file */
}

	  
